model: number threads from 0, not 1
[c11tester.git] / model.cc
index 1f13de7f1e5ac4c74cea012bc096f0a23d666cfe..c829312166033342c8c8d077f0ec2acd63c33230 100644 (file)
--- a/model.cc
+++ b/model.cc
@@ -1,15 +1,20 @@
 #include <stdio.h>
 
 #include "model.h"
+#include "action.h"
+#include "tree.h"
 #include "schedule.h"
 #include "common.h"
 
+#define INITIAL_THREAD_ID      0
+
 ModelChecker *model;
 
 ModelChecker::ModelChecker()
 {
-       /* First thread created (system_thread) will have id 1 */
-       this->used_thread_id = 0;
+       /* First thread created will have id INITIAL_THREAD_ID */
+       this->next_thread_id = INITIAL_THREAD_ID;
+       used_sequence_numbers = 0;
        /* Initialize default scheduler */
        this->scheduler = new Scheduler();
 
@@ -30,14 +35,29 @@ ModelChecker::~ModelChecker()
        delete rootNode;
 }
 
-int ModelChecker::get_next_id()
+void ModelChecker::reset_to_initial_state()
+{
+       DEBUG("+++ Resetting to initial state +++\n");
+       std::map<int, class Thread *>::iterator it;
+       for (it = thread_map.begin(); it != thread_map.end(); it++)
+               delete (*it).second;
+       thread_map.clear();
+       action_trace = new action_list_t();
+       currentNode = rootNode;
+       current_action = NULL;
+       next_thread_id = INITIAL_THREAD_ID;
+       used_sequence_numbers = 0;
+       /* scheduler reset ? */
+}
+
+thread_id_t ModelChecker::get_next_id()
 {
-       return ++used_thread_id;
+       return next_thread_id++;
 }
 
-void ModelChecker::add_system_thread(Thread *t)
+int ModelChecker::get_next_seq_num()
 {
-       this->system_thread = t;
+       return ++used_sequence_numbers;
 }
 
 Thread * ModelChecker::schedule_next_thread()
@@ -45,7 +65,7 @@ Thread * ModelChecker::schedule_next_thread()
        Thread *t;
        if (nextThread == THREAD_ID_T_NONE)
                return NULL;
-       t = thread_map[nextThread];
+       t = thread_map[id_to_int(nextThread)];
        if (t == NULL)
                DEBUG("*** error: thread not in thread_map: id = %d\n", nextThread);
        return t;
@@ -97,6 +117,7 @@ thread_id_t ModelChecker::advance_backtracking_state()
 bool ModelChecker::next_execution()
 {
        num_executions++;
+       print_summary();
        if ((exploring = model->get_next_backtrack()) == NULL)
                return false;
        model->reset_to_initial_state();
@@ -106,9 +127,7 @@ bool ModelChecker::next_execution()
 
 ModelAction * ModelChecker::get_last_conflict(ModelAction *act)
 {
-       void *loc = act->get_location();
        action_type type = act->get_type();
-       thread_id_t id = act->get_tid();
 
        switch (type) {
                case THREAD_CREATE:
@@ -124,14 +143,8 @@ ModelAction * ModelChecker::get_last_conflict(ModelAction *act)
        action_list_t::reverse_iterator rit;
        for (rit = action_trace->rbegin(); rit != action_trace->rend(); rit++) {
                ModelAction *prev = *rit;
-               if (prev->get_location() != loc)
-                       continue;
-               if (type == ATOMIC_READ && prev->get_type() != ATOMIC_WRITE)
-                       continue;
-               /* Conflict from the same thread is not really a conflict */
-               if (id == prev->get_tid())
-                       return NULL;
-               return prev;
+               if (act->is_dependent(prev))
+                       return prev;
        }
        return NULL;
 }
@@ -154,15 +167,27 @@ void ModelChecker::set_backtracking(ModelAction *act)
        if (node->setBacktrack(act->get_tid()) != 0)
                return;
 
-       printf("Setting backtrack: conflict = %d, instead tid = %d\n",
+       DEBUG("Setting backtrack: conflict = %d, instead tid = %d\n",
                        prev->get_tid(), act->get_tid());
-       prev->print();
-       act->print();
+       if (DBG_ENABLED()) {
+               prev->print();
+               act->print();
+       }
 
        Backtrack *back = new Backtrack(prev, action_trace);
        backtrack_list.push_back(back);
 }
 
+Backtrack * ModelChecker::get_next_backtrack()
+{
+       Backtrack *next;
+       if (backtrack_list.empty())
+               return NULL;
+       next = backtrack_list.back();
+       backtrack_list.pop_back();
+       return next;
+}
+
 void ModelChecker::check_current_action(void)
 {
        ModelAction *next = this->current_action;
@@ -171,6 +196,7 @@ void ModelChecker::check_current_action(void)
                DEBUG("trying to push NULL action...\n");
                return;
        }
+       current_action = NULL;
        nextThread = advance_backtracking_state();
        next->set_node(currentNode);
        set_backtracking(next);
@@ -178,7 +204,7 @@ void ModelChecker::check_current_action(void)
        this->action_trace->push_back(next);
 }
 
-void ModelChecker::print_trace(void)
+void ModelChecker::print_summary(void)
 {
        action_list_t::iterator it;
 
@@ -189,7 +215,7 @@ void ModelChecker::print_trace(void)
 
        scheduler->print();
 
-       printf("\nTrace:\n\n");
+       printf("Trace:\n\n");
 
        for (it = action_trace->begin(); it != action_trace->end(); it++) {
                DBG();
@@ -200,20 +226,25 @@ void ModelChecker::print_trace(void)
 
 int ModelChecker::add_thread(Thread *t)
 {
-       thread_map[t->get_id()] = t;
+       thread_map[id_to_int(t->get_id())] = t;
+       scheduler->add_thread(t);
        return 0;
 }
 
+void ModelChecker::remove_thread(Thread *t)
+{
+       scheduler->remove_thread(t);
+}
+
 int ModelChecker::switch_to_master(ModelAction *act)
 {
-       Thread *old, *next;
+       Thread *old;
 
        DBG();
        old = thread_current();
        set_current_action(act);
        old->set_state(THREAD_READY);
-       next = system_thread;
-       return old->swap(next);
+       return Thread::swap(old, get_system_context());
 }
 
 ModelAction::ModelAction(action_type_t type, memory_order order, void *loc, int value)
@@ -226,6 +257,63 @@ ModelAction::ModelAction(action_type_t type, memory_order order, void *loc, int
        act->location = loc;
        act->tid = t->get_id();
        act->value = value;
+       act->seq_number = model->get_next_seq_num();
+}
+
+bool ModelAction::is_read()
+{
+       return type == ATOMIC_READ;
+}
+
+bool ModelAction::is_write()
+{
+       return type == ATOMIC_WRITE;
+}
+
+bool ModelAction::is_acquire()
+{
+       switch (order) {
+       case memory_order_acquire:
+       case memory_order_acq_rel:
+       case memory_order_seq_cst:
+               return true;
+       default:
+               return false;
+       }
+}
+
+bool ModelAction::is_release()
+{
+       switch (order) {
+       case memory_order_release:
+       case memory_order_acq_rel:
+       case memory_order_seq_cst:
+               return true;
+       default:
+               return false;
+       }
+}
+
+bool ModelAction::same_var(ModelAction *act)
+{
+       return location == act->location;
+}
+
+bool ModelAction::same_thread(ModelAction *act)
+{
+       return tid == act->tid;
+}
+
+bool ModelAction::is_dependent(ModelAction *act)
+{
+       if (!is_read() && !is_write())
+               return false;
+       if (!act->is_read() && !act->is_write())
+               return false;
+       if (same_var(act) && !same_thread(act) &&
+                       (is_write() || act->is_write()))
+               return true;
+       return false;
 }
 
 void ModelAction::print(void)
@@ -251,5 +339,6 @@ void ModelAction::print(void)
                type_str = "unknown type";
        }
 
-       printf("Thread: %d\tAction: %s\tMO: %d\tLoc: %#014zx\tValue: %d\n", tid, type_str, order, (size_t)location, value);
+       printf("(%4d) Thread: %d\tAction: %s\tMO: %d\tLoc: %14p\tValue: %d\n",
+                       seq_number, id_to_int(tid), type_str, order, location, value);
 }